package rbas.ranksearch.app;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.BufferedReader;
import java.io.FileReader;

import rbas.ranksearch.search.RankSearch;
import rbas.ranksearch.sem.MixtureOfSemsEstimator;
import edu.cmu.tetrad.data.RectangularDataSet;
import edu.cmu.tetrad.data.ColtDataSet;
import edu.cmu.tetrad.data.ContinuousVariable;
import edu.cmu.tetrad.graph.*;

/**
 * RankSearchApp: command line interface for running RankSearch.
 *
 * @author Ricardo Silva, Fall 2004 - Fall 2006
 * @version $Revision: 1.1 $ $Date: 2006/09/09 19:42:51 $
 */

public final class RankSearchApp {

    private RectangularDataSet dataSet, testData;
    private int numMixtures;

    public RankSearchApp(String fileName, int numMixtures) {
        loadData(fileName);
        this.numMixtures = numMixtures;
    }

    public static void main(final String[] args) {
        if (args.length != 2) {
            System.out.println("RankSearchApp usage:");
            System.out.println("   RankSearchApp <data file name> <number of mixture components>");
            System.exit(0);
        }

        try {
            RankSearchApp rs = new RankSearchApp(args[0], Integer.parseInt(args[1]));

            java.util.Date date1 = new java.util.Date();
            rs.search();
            System.out.println("** END OF SEARCH **");
            java.util.Date date2 = new java.util.Date();
            System.out.println("Started at " + date1.toString() +  
                               ",finished at " + date2.toString());
        } catch (NumberFormatException e) {
            System.out.println("Invalid arguments! Only integers are accepted.");
            throw new RuntimeException(e);
        }
    }

    public void loadData(String name) {
        try {
            System.out.println("Loading datasets " + name);
            this.dataSet = loadDataSet(name + ".dat");
            this.testData = loadDataSet(name + ".tst");
        } catch (java.io.FileNotFoundException e) {
            throw new java.lang.RuntimeException(e);
        } catch (java.io.IOException e) {
            throw new java.lang.RuntimeException(e);
        }
    }

    private static RectangularDataSet loadDataSet(String fileName) throws java.io.IOException {
        BufferedReader in = new BufferedReader(new FileReader(fileName));
        int lineCount = 0;

        String line;
        do {
            line = in.readLine();
            if (line != null)
               line = line.trim();
            lineCount++;
        } while (line != null && line.equals(""));

        if (line == null) {
            System.out.println("Invalid dataset: " + fileName + "!");
            System.exit(0);
        }

        String columnNames[] = line.split(" +|\t+"); //header
        List<List<Double>> data = new ArrayList<List<Double>>();
        for (int i = 0; i < columnNames.length; i++)
             data.add(new ArrayList<Double>());
        while (true) {
            line = in.readLine();
            if (line == null)
                break;
            line = line.trim();
            lineCount++;
            if (line.equals(""))
                continue;
            String row[] = line.split(" +|\t+");
            if (row.length != columnNames.length) {
                System.out.println("Number of columns in line " + lineCount + " does not match number of variables!");
                System.exit(0);
            }
            int cCount = 0;
            for (List<Double> dColumn : data)
                dColumn.add(new Double(row[cCount++]));
        }

        List<Node> columnNodes = new ArrayList<Node>();
        for (int i = 0; i < columnNames.length; i++)
             columnNodes.add(new ContinuousVariable(columnNames[i]));
        RectangularDataSet output = new ColtDataSet(data.get(0).size(), columnNodes);
        int rows = data.get(0).size(), cols = columnNames.length;
        for (int j = 0; j < cols; j++) {
            List<Double> dColumn = data.get(j);
            for (int i = 0; i < rows; i++)
                output.setDouble(i, j, (double) dColumn.get(i));
        }
        return output;
    }


    public void search() {
        RankSearch mimScore = new RankSearch(this.dataSet, this.numMixtures);
        Graph mimGraph = mimScore.search();
        System.out.println("-----------------------------------------------------");
        System.out.println("Estimated graph:");
        System.out.println();
        System.out.println(mimGraph.toString());
        scoreGraph(mimGraph);
    }

    private void scoreGraph(Graph graph) {
        Graph completedGraph = new EdgeListGraph(graph);
        List<Edge> bidirectedEdges = new ArrayList<Edge>();
        for (Iterator it = completedGraph.getEdges().iterator(); it.hasNext();) {
            Edge next = (Edge) it.next();
            if (next.getEndpoint1() == Endpoint.ARROW && next.getEndpoint2() == Endpoint.ARROW) {
                bidirectedEdges.add(next);
                Node newL = new GraphNode("___L" + completedGraph.getNodes().size()); newL.setNodeType(NodeType.LATENT);
                completedGraph.addNode(newL);
                completedGraph.addDirectedEdge(newL, next.getNode1()); completedGraph.addDirectedEdge(newL, next.getNode2());
            }
        }
        completedGraph.removeEdges(bidirectedEdges);
        graph = completedGraph;
        System.out.println("** COMPLETED GRAPH");
        System.out.println(graph.toString());

        if (this.testData != null)
            System.out.println("  >> Average likelihood (test): " +
                               scoreModel(graph, this.dataSet, this.testData) / this.testData.getNumRows());

    }

    private double scoreModel(Graph graph, RectangularDataSet dataSet, RectangularDataSet testSet) {
        double trainScore = -Double.MAX_VALUE;
        MixtureOfSemsEstimator bestEstTrain = null;
        for (int c = 0; c < 10; c++) {
            MixtureOfSemsEstimator estTrain = new MixtureOfSemsEstimator(this.dataSet, graph, this.numMixtures);
            estTrain.estimate();
            if (estTrain.logLikelihood() > trainScore) {
                trainScore = estTrain.logLikelihood();
                bestEstTrain = estTrain;
                System.out.println("***score = " + " (ll = " + estTrain.logLikelihood() + ")");
            } else
                System.out.println("///score = " + " (ll = " + estTrain.logLikelihood() + ")");
            System.out.println("---average train score =  (ll = " + estTrain.logLikelihood() /
                                                                    dataSet.getNumRows() + ")");
        }
        List observables = new ArrayList();
        for (Iterator it = graph.getNodes().iterator(); it.hasNext();) {
            Node next = (Node) it.next();
            if (next.getNodeType() == NodeType.MEASURED)
                observables.add(next);
        }
        System.out.println("-- best train likelihood: "+ bestEstTrain.logLikelihood() / dataSet.getNumRows());
        return bestEstTrain.logLikelihoodEval(testSet);
    }

}
